UNIT SPR_XMS;

{
 Ergnzung zu SPRITE.TPU,
 erweitert den Funktionsumfang um XMS-Routinen.

 (c) 1997 by Jan Strter

 Version: 1.5
}

{ ************************************************************************ }
INTERFACE
{ ************************************************************************ }

USES
    sprite;

CONST
      xms_NoError = 0;                       { Fehlercodes }
      xms_OutOfMemError = 1;                 { Kein Speicher mehr da }
      xms_IOError = 2;                       { Datei-Fehler }
      xms_WrongFormatError = 3;              { Falsches Format }
      xms_AnyError = 4;                      { Unbekannter Fehler }
      xms_Error: byte = xms_NoError;         { Standardmig kein Fehler }
      Rmem_Max : longint = 3*1024*1024-70000;{ Maximale gre des XMS }

VAR
{    xms_virtualvseg: word;}                 { Zeiger auf Bildschirmpuffer, fr Assemblerteil }
    xms_vbuf: pointer;
    GDT_Off : ARRAY[1..6] OF byte;

CONST GDT : ARRAY[1..16] OF byte =(
  $00,$00,$00,$00,$00,$00,$00,$00,        {GDT Eintrag 0 (null segment)}
  $FF,$FF,$00,$00,$00,$92,$CF,$FF);       {GDT Eintrag 1 (seg 0, limit 4GB)}

PROCEDURE memory_checks(minmain,minxms : word);
          { Prft, ob genug Speicher, wenn nicht -> Programmende }
PROCEDURE enable_Realmem(Min : word);
          { Belegt Speicher, schaltet in V86-Modus }
PROCEDURE Exit_Rmem;
          { Gibt allen belegten Speicher wieder frei - WICHTIG! }
FUNCTION  Rgetmem(VAR rpos : longint;rsize : longint) : boolean;
          { Belegt Speicher im XMS }
FUNCTION  RFreeMem(rsize: longint): boolean;
          { Gibt Speicher wieder frei }
PROCEDURE Rmem_Lesen(quelle:longint; ziel:pointer;laenge:word);
          { Liest aus dem XMS }
PROCEDURE Rmem_write(quelle:pointer;ziel:longint;laenge:word);
          { Schreibt in den XMS }
PROCEDURE xms_sprites_init(Vbuffer: pointer; konv_speicher,Xms_speicher: longint);
          { Allokiert Speicher und belegt xms_virtualVSeg vor }
PROCEDURE xms_sprites_done;
          { Gibt Speicher wieder frei, MUSS aufgerufen werden! }
PROCEDURE xms_push_sprite(qsprite: spritetyp; VAR zsprite: spritetyp);
          { Kopiert ein Sprite in den XMS (flat-mode) }
PROCEDURE xms_pop_sprite(VAR zsprite: spritetyp);
          { Entfernt ein Sprite aus dem XMS }
PROCEDURE xms_putsprite_fast(x,y: word; qsprite: spritetyp);
          { Putsprite_fast, halt fr XMS... }
PROCEDURE xms_putbackground(adr: pointer);
          { Zeichnet Hintergrund-Sprite (320*200) aus dem XMS, erwartet NICHT
            Sprite, sondern nur Adresse (z.B. THESPRITE.ADR)! }
PROCEDURE xms_putsprite_medium(x,y: integer; qsprite: spritetyp);
          { Selbstklrend, wenn man SPRITE.TPU kennt... }
FUNCTION xms_putsprite_exact(x,y: integer; zsprite: spritetyp): boolean;
          { s.o. }
PROCEDURE xms_putsprite_shadow(x,y: integer; VAR zsprite: spritetyp);
          { Zeichnet das Sprite als Schatten. (Erst MAKESHADOWPAL aufrufen!)}
PROCEDURE xms_putsprite_trans(x,y: integer; VAR zsprite: spritetyp);
          { Zeichnet Sprite transparent. (Erste LOADTRANSPAL oder MAKETRANSPAL aufrufen!) }
FUNCTION xms_spriteoverlap_exact(x1,y1: integer; sprite1: spritetyp; x2,y2: integer; sprite2: spritetyp): boolean;
          { Stellt fest, ob sich zwei Sprites berschneiden. }

          { Anmerkung:
            der XMS-Speicher wird von der Unit der Einfachheit halber
            als 'Stack' behandelt (daher PUSH und POP). Die Sprites
            mssen daher in umgekehrter Reihenfolge wieder entfernt
            werden (POP_SPRITE_XMS), wie sie mit PUSH_SPRITE_XMS in den
            'XMS-Stack' 'gepusht' wurden.
          }
{ ************************************************************************ }
IMPLEMENTATION
{ ************************************************************************ }

TYPE XMSHandle = word;

     XMS_Copyblock = RECORD     { Wird fr die Kopier-Routinen bentigt }
       Size     : longint;
       Q_Handle : Word;
       Q_Offset : pointer;
       Z_Handle : Word;
       Z_Offset : pointer;
     END;

VAR XMS_Vorhanden   : boolean;  { TRUE, wenn XMS vorhanden ist }
    XMST            : pointer;  { Treiber - Einsprungadresse }
    XMS_Version     : word;     { Die Version des XMS-Treibers }
    XC              : XMS_Copyblock;
    xms_frei : longint;
    error : integer;
    My_XmsHandle : XmsHandle;
    Xms_startposi : longint;
    Old_ExitprocRmem : pointer;


PROCEDURE xms_sprites_init;
BEGIN
     memory_checks(konv_speicher,xms_speicher);
     enable_realmem(xms_speicher);
     xms_vbuf := vbuffer;
END;

PROCEDURE xms_sprites_done;
BEGIN
     exit_rmem;
END;

PROCEDURE xms_push_sprite;
VAR
   startpos: longint;
   fehler: boolean;
BEGIN
     fehler:=rgetmem(startpos,longint(qsprite.dtx*qsprite.dty));
     IF fehler <> true THEN BEGIN
        xms_error := xms_OutOfMemError;
        exit;
     END;
     rmem_write(qsprite.adr,startpos,qsprite.dtx*qsprite.dty);
     WITH zsprite DO BEGIN
          adr := pointer(startpos);
          dtx := qsprite.dtx;
          dty := qsprite.dty;
     END;
END;

PROCEDURE xms_pop_sprite;
VAR
   fehler: boolean;
BEGIN
     fehler := rfreemem(zsprite.dtx*zsprite.dty);
     IF fehler <> true THEN BEGIN
        xms_error := xms_AnyError;
        exit;
     END;
     WITH zsprite DO BEGIN
          adr := NIL;
          dtx := 0;
          dty := 0;
     END;
END;

FUNCTION XMS_free  : longint;
VAR xms_in_kb : word;
    xms_long: longint;
BEGIN;
  ASM
    mov ax,0800h                 { 8 = Freien Speicher Ermitteln }
    call dword ptr [XMST]
    mov xms_in_kb,dx
  END;
  xms_long := xms_in_kb;
  XMS_free := xms_long * 1024;
END;

FUNCTION Getmem_XMS(VAR H : XMSHandle; Size : longint) : byte;
VAR bsize : word;
    Fresult : byte;
    xmsh : word;
BEGIN;
  bsize := (size DIV 1024) + 1;
  ASM
    mov ax,0900h                 { 9 = Speicherbereich allocieren }
    mov dx,bsize
    call dword ptr [XMST]
    cmp ax,1
    jne @Fehler_GetmemXms
    mov xmsh,dx
    mov Fresult,0
    jmp @Ende_GetmemXms
@Fehler_GetmemXMS:
    mov Fresult,bl
@Ende_GetmemXms:
  END;
  h := xmsh;
  Getmem_Xms := Fresult;
END;

FUNCTION Freemem_XMS(H : XMSHandle) : byte;
VAR fresult : byte;
BEGIN;
  ASM                            { A = Speicherbereich deallocieren }
    mov ax,0a00h
    mov dx,h
    call dword ptr [XMST]
    cmp ax,1
    jne @Fehler_FreememXms
    mov Fresult,0
    jmp @Ende_FreememXms
@Fehler_FreememXms:
    mov Fresult,bl
@Ende_FreememXms:
  END;
END;

PROCEDURE Check_for_XMS; assembler;
  ASM
    mov ax,4300h              { Prfen, ob Treiber Installiert }
    int 2Fh
    cmp al,80h
    jne @Kein_XMSTreiber
    mov ax,4310h              { Einsprungadresse des Treibers ermitteln }
    int 2Fh
    mov word ptr XMST + 2,es
    mov word ptr XMST + 0,bx
    xor ax,ax                 { Versionsnummer ermitteln }
    call dword ptr [XMST]
    cmp  ax,0200h
    jb   @Kein_XMSTreiber     { Wenn Version < 2.0 dann Abbrechen ! }
    mov  XMS_Version,ax
    mov  XMS_Vorhanden,0
@Kein_XMSTreiber:
    mov XMS_Vorhanden,1
@Ende_XMS_Check:
END;

FUNCTION XMS_lock(H : XMSHandle) : longint; assembler;
ASM;
  mov ax,0c00h
  mov dx,h
  call dword ptr [XMST]
  mov ax,bx
END;

PROCEDURE XMS_unlock(H : XMSHandle); assembler;
ASM;
  mov ax,0d00h
  mov dx,h
  call dword ptr [XMST]
END;

PROCEDURE XMS_Enable_A20; assembler;
ASM
  mov ax,0500h
  call dword ptr [XMST]
END;

PROCEDURE XMS_Disable_A20; assembler;
ASM
  mov ax,0600h
  call dword ptr [XMST]
END;


const MByte1: longint = $100000;

VAR Offs,Segm : word;
    Rmemposi  : longint;

{$l rmemasm.obj}
FUNCTION xms_spriteoverlap_asm(x1,y1: integer; sprite1: pointer; x2,y2: integer; sprite2: pointer; restx1,restx2: longint;
                               startx1,startx2,starty1,starty2,dtx1,dtx2,width,height: word):byte; FAR; EXTERNAL;
PROCEDURE xms_putbackground(adr: pointer); EXTERNAL;
PROCEDURE xms_putsprite_fast_asm(x,y: word; adr: pointer; dtx,dty,dwords,bytes: word); FAR; EXTERNAL;
PROCEDURE xms_putsprite_med(adr: pointer; x,y: word; c_r,c_l,c_u,c_d,rest,dtx,dty,wdt,dwords,bytes: word); FAR; EXTERNAL;
PROCEDURE xms_putsprite(adr: pointer; x,y,c_l,c_r,c_u,c_d,rest_q,rest_z,dtx,dty: word); FAR; EXTERNAL;
PROCEDURE putsprite_shadow_asm(adr: pointer; x,y,c_l,c_r,c_u,c_d,rest_q,rest_z,dtx,dty,tab_seg,tab_ofs: word); FAR; EXTERNAL;
PROCEDURE putsprite_trans_asm(adr: pointer; x,y,c_l,c_r,c_u,c_d,rest_q,rest_z,dtx,dty,transpal: word); FAR; EXTERNAL;
PROCEDURE mem_write(q:longint;zl,zh,l:word); FAR; EXTERNAL;
{
 *************************************************************************
 ***                                                                   ***
 ***       Kopiert einen Block aus dem Hauptspeicher ins RMEM          ***
 ***                                                                   ***
 *************************************************************************
}

PROCEDURE mem_lesen(q:longint;zl,zh,l:word); FAR; EXTERNAL;
{
 *************************************************************************
 ***                                                                   ***
 ***      Kopiert einen Block aus dem RMEM in den Hauptspeicher        ***
 ***                                                                   ***
 *************************************************************************
}

PROCEDURE Enable_4Giga; FAR; EXTERNAL;
{
 *************************************************************************
 ***                                                                   ***
 ***            Schaltet den Processor ins Flat - Model                ***
 ***                                                                   ***
 *************************************************************************
}

FUNCTION multitasker_aktiv : boolean; FAR; EXTERNAL;
{
 *************************************************************************
 ***                                                                   ***
 ***   Prft, ob ein Multitasker wie QEMM oder EMM386 aktiv ist        ***
 ***                                                                   ***
 *************************************************************************
}

PROCEDURE xms_putsprite_fast;
VAR
   dwords,bytes: word;
BEGIN
     dwords := qsprite.dtx DIV 4;
     bytes := qsprite.dtx MOD 4;
     WITH qsprite DO xms_putsprite_fast_asm(word(x),word(y),adr,dtx,dty,dwords,bytes);
END;

FUNCTION xms_putsprite_exact;
VAR
   startx,starty,endx: word;
   clip_dn,clip_up,clip_lt,clip_rt: word;
   ofs_spr: word;
   seg_spr: word;
   width: word;
   restx: word;
   dtx,dty:word;
BEGIN
     xms_putsprite_exact:=false;
     IF ((x>319) OR (y>199)) OR ((x+zsprite.dtx<=0) OR (y+zsprite.dty<=0)) THEN exit;
     xms_putsprite_exact:=true;
     width:=zsprite.dtx;
     ASM                        { Von Hand ist es 'nen bichen schneller,}
        xor ax,ax               { weil TP vor jedem MOV noch ein berflssiges}
        mov clip_rt,ax          { XOR AX,AX einfgt. }
        mov clip_lt,ax
        mov clip_dn,ax
        mov clip_up,ax
     END;
     endx:=x+zsprite.dtx;
     IF x<0 THEN
     BEGIN
          startx:=0;
          clip_lt:=-x;
     END
     ELSE BEGIN
          startx:=x;
          IF x+zsprite.dtx>320 THEN
          BEGIN
               endx:=320;
               clip_rt:=(x+zsprite.dtx-320);
          END;
     END;
     IF y<0 THEN
     BEGIN
          starty:=0;
          clip_up:=-y;
     END
     ELSE BEGIN
          starty:=y;
          IF y+zsprite.dty>200 THEN
          BEGIN
               clip_dn:=(y+zsprite.dty-200);
          END;
     END;
     seg_spr:=seg(zsprite.adr^);
     ofs_spr:=ofs(zsprite.adr^);
     restx:=320-endx;
     dtx:=zsprite.dtx-clip_rt-clip_lt;
     dty:=zsprite.dty-clip_up-clip_dn;
     xms_putsprite(zsprite.adr,startx,starty,clip_lt,clip_rt,clip_up,clip_dn,restx,zsprite.dtx,dty,dtx);
END;

PROCEDURE xms_putsprite_shadow;
VAR
   startx,starty,endx: word;
   clip_dn,clip_up,clip_lt,clip_rt: word;
   ofs_spr: word;
   seg_spr: word;
   width: word;
   restx: word;
   dtx,dty:word;
   data_seg: word;
   shadow_ofs: word;
BEGIN
     IF ((x>319) OR (y>199)) OR ((x+zsprite.dtx<=0) OR (y+zsprite.dty<=0)) THEN exit;
     width:=zsprite.dtx;
     ASM                        { Von Hand ist es 'nen bichen schneller,}
        xor ax,ax               { weil TP vor jedem MOV noch ein berflssiges}
        mov clip_rt,ax          { XOR AX,AX einfgt. }
        mov clip_lt,ax
        mov clip_dn,ax
        mov clip_up,ax
     END;
     endx:=x+zsprite.dtx;
     IF x<0 THEN
     BEGIN
          startx:=0;
          clip_lt:=-x;
     END
     ELSE BEGIN
          startx:=x;
          IF x+zsprite.dtx>320 THEN
          BEGIN
               endx:=320;
               clip_rt:=(x+zsprite.dtx-320);
          END;
     END;
     IF y<0 THEN
     BEGIN
          starty:=0;
          clip_up:=-y;
     END
     ELSE BEGIN
          starty:=y;
          IF y+zsprite.dty>200 THEN
          BEGIN
               clip_dn:=(y+zsprite.dty-200);
          END;
     END;
     seg_spr:=seg(zsprite.adr^);
     ofs_spr:=ofs(zsprite.adr^);
     shadow_ofs:=ofs(shadowpal);
     restx:=320-endx;
     dtx:=zsprite.dtx-clip_rt-clip_lt;
     dty:=zsprite.dty-clip_up-clip_dn;
{     xms_putsprite(zsprite.adr,startx,starty,clip_lt,clip_rt,clip_up,clip_dn,restx,zsprite.dtx,dty,dtx);}
     putsprite_shadow_asm(zsprite.adr,startx,starty,clip_lt,clip_rt,clip_up,clip_dn,restx,zsprite.dtx,dty,dtx,
                          seg(shadowpal),ofs(shadowpal));
END;                                    {Bis zum nchsten Mal, tsch.}

PROCEDURE xms_putsprite_trans;
VAR
   startx,starty,endx: word;
   clip_dn,clip_up,clip_lt,clip_rt: word;
   ofs_spr: word;
   seg_spr: word;
   width: word;
   restx: word;
   dtx,dty:word;
BEGIN
     IF ((x>319) OR (y>199)) OR ((x+zsprite.dtx<=0) OR (y+zsprite.dty<=0)) THEN exit;
     width:=zsprite.dtx;
     ASM                        { Von Hand ist es 'nen bichen schneller,}
        xor ax,ax               { weil TP vor jedem MOV noch ein berflssiges}
        mov clip_rt,ax          { XOR AX,AX einfgt. }
        mov clip_lt,ax
        mov clip_dn,ax
        mov clip_up,ax
     END;
     endx:=x+zsprite.dtx;
     IF x<0 THEN
     BEGIN
          startx:=0;
          clip_lt:=-x;
     END
     ELSE BEGIN
          startx:=x;
          IF x+zsprite.dtx>320 THEN
          BEGIN
               endx:=320;
               clip_rt:=(x+zsprite.dtx-320);
          END;
     END;
     IF y<0 THEN
     BEGIN
          starty:=0;
          clip_up:=-y;
     END
     ELSE BEGIN
          starty:=y;
          IF y+zsprite.dty>200 THEN
          BEGIN
               clip_dn:=(y+zsprite.dty-200);
          END;
     END;
     seg_spr:=seg(zsprite.adr^);
     ofs_spr:=ofs(zsprite.adr^);
     restx:=320-endx;
     dtx:=zsprite.dtx-clip_rt-clip_lt;
     dty:=zsprite.dty-clip_up-clip_dn;
     putsprite_trans_asm(zsprite.adr,startx,starty,clip_lt,clip_rt,clip_up,clip_dn,restx,zsprite.dtx,dty,dtx,
                          seg(transpal^));
END;                                    {Bis zum nchsten Mal, tsch.}

PROCEDURE xms_putsprite_medium;
VAR
   c_r,c_l,c_u,c_d:word;        { Clipping an den Kanten }
   restx: word;
   anz_dwords: word;            { Anzahl von DWords, die kopiert werden }
   anz_bytes: word;             { Anzahl Bytes, die kopiert werden }
   dtx,dty: word;               { Breite und Hhe, abzglich des Clippings }
   restx_spr: word;
   startx,endx: word;           { Koordinaten am Bildschirm }
   starty,endy: word;
BEGIN
     { Prfen, ob berhaupt sichtbar, wenn nicht, dann sofort raus }
     IF (x>319) OR (y>199) THEN exit;
     IF (x+qsprite.dtx<=0) OR (y+qsprite.dty<=0) THEN exit;
     { Clipping berechnen }
     c_r := 0; c_l := 0; c_u := 0; c_d := 0; { Voreinstellung: 0 }
     IF x < 0 THEN BEGIN
        c_l := -x;
        startx:=0;
     END ELSE startx:=x;
     IF y < 0 THEN BEGIN
        c_u := -y;
        starty:=0;
     END ELSE starty:=y;
     IF x+qsprite.dtx > 320 THEN BEGIN
        c_r := x+qsprite.dtx-320;
        endx:=320;
     END ELSE endx:=x+qsprite.dtx;
     IF y+qsprite.dty > 200 THEN BEGIN
        c_d := y+qsprite.dty-200;
        endy := 200;
     END ELSE endy:=y+qsprite.dty;
     { Aufspaltung der Daten in DWord- und Bytehppchen }
     dtx := qsprite.dtx-c_r-c_l;
     dty := qsprite.dty-c_u-c_d;
     anz_dwords := dtx DIV 4;
     anz_bytes := dtx mod 4;
     { Diese Werte mssen nach jeder Zeile dazuaddiert werden, um ein
       korrektes Bild zu erhalten }
     restx:=320-endx;
     xms_putsprite_med(qsprite.adr,startx,starty,c_r,c_l,c_u,c_d,restx,dtx,dty,qsprite.dtx,anz_dwords,anz_bytes);
END;                                    { ...und tsch! }

FUNCTION xms_spriteoverlap_exact;
VAR
   start_x_1,start_x_2,start_y_1,start_y_2: word;
   end_x_1,end_x_2,end_y_1,end_y_2: word;
   breite,hoehe: word;
   gotcha: byte;
   rest_x_1,rest_x_2: word;
BEGIN
     xms_spriteoverlap_exact:=false;

     { Haben wir berhaupt eine berschneidung? }
     IF spriteoverlap_fast(x1,y1,sprite1,x2,y2,sprite2) = false THEN exit;

     { Ja, also bestimmen wir berschneidungsbreich }

     { Fall 1: Erstes Sprite befindet sich in X-Richtung im 2. Sprite (links)}
     IF ((x1>=x2) AND (x1<=x2+sprite2.dtx-1)) THEN BEGIN
        start_x_1:=x1;
        start_x_2:=x1;
     { Fall 2: 2. Sprite befindet sich in X-Richtung im 1. Sprite (links)}
     END ELSE IF ((x2>=x1) AND (x2<=x1+sprite1.dtx-1)) THEN BEGIN
        start_x_1:=x2;
        start_x_2:=x2;
     END;

     { Fall 1: Erstes Sprite befindet sich in X-Richtung im 2. Sprite (rechts)}
     IF ((x1+sprite1.dtx-1>=x2) AND (x1+sprite1.dtx-1<=x2+sprite2.dtx-1)) THEN BEGIN
        end_x_1:=x1+sprite1.dtx-1;
        end_x_2:=x1+sprite1.dtx-1;
     { Fall 2: 2. Sprite befindet sich in X-Richtung im 1. Sprite (rechts)}
     END ELSE IF ((x2+sprite2.dtx-1>=x1) AND (x2+sprite2.dtx-1<=x1+sprite1.dtx-1)) THEN BEGIN
        end_x_1:=x2+sprite2.dtx-1;
        end_x_2:=x2+sprite2.dtx-1;
     END;

     { Fall 1: 1. Sprite befindet sich in Y-Richtung im 2. Sprite (oben)}
     IF ((y1>=y2) AND (y1<=y2+sprite2.dty-1)) THEN BEGIN
        start_y_1:=y1;
        start_y_2:=y1;
     { Fall 2: 2. Sprite befindet sich in Y-Richtung im 1. Sprite (oben)}
     END ELSE IF ((y2>=y1) AND (y2<=y1+sprite1.dty-1)) THEN BEGIN
        start_y_1:=y2;
        start_y_2:=y2;
     END;

     { Fall 1: 1. Sprite befindet sich in Y-Richtung im 2. Sprite (unten)}
     IF ((y1+sprite1.dty-1>=y2) AND (y1+sprite1.dty-1<=y2+sprite2.dty-1)) THEN BEGIN
        end_y_1:=y1+sprite1.dty-1;
        end_y_2:=y1+sprite1.dty-1;
     { Fall 2: 2. Sprite befindet sich in Y-Richtung im 1. Sprite (unten)}
     END ELSE IF ((y2+sprite2.dty-1>=y1) AND (y2+sprite2.dty-1<=y1+sprite1.dty-1)) THEN BEGIN
        end_y_1:=y2+sprite2.dty-1;
        end_y_2:=y2+sprite2.dty-1;
     END;

     { Wir wollen keine absoluten Koordinaten, sondern die innerhalb der
       Sprites }
     dec(start_x_1,x1);
     dec(start_x_2,x2);
     dec(start_y_1,y1);
     dec(start_y_2,y2);
     dec(end_x_1,x1);
     dec(end_x_2,x2);
     dec(end_y_1,y1);
     dec(end_y_2,y2);

     { Breite und Hhe des berschneidungsbereiches bestimmen }
     breite:=end_x_1-start_x_1;
     hoehe:=end_y_1-start_y_1;

     { Um diese Werte mssen die Offsets der Sprites nach jeder Zeile
       erhht werden }
     rest_x_1:=sprite1.dtx-breite;
     rest_x_2:=sprite2.dtx-breite;

     { Wenn der Bereich kleiner Null ist, liest der Rechner irgendwelchen
       Mll aus dem Speicher, das mu verhindert werden! }

     IF (breite < 1) OR (hoehe < 1) THEN exit;

     gotcha := 0;

     gotcha:=xms_spriteoverlap_asm(x1,y1,sprite1.adr,x2,y2,sprite2.adr,longint(rest_x_1),longint(rest_x_2),start_x_1,
                                   start_x_2,start_y_1,start_y_2,sprite1.dtx,sprite2.dtx,breite,hoehe);

     IF gotcha = 1 THEN xms_spriteoverlap_exact:=true;
END;                            {...und tschss! }

PROCEDURE Rmem_Lesen(quelle:longint; ziel:pointer;laenge:word);
{
 *************************************************************************
 ***                                                                   ***
 ***      Kopiert einen Block aus dem RMEM in den Hauptspeicher        ***
 ***                                                                   ***
 *************************************************************************
}
BEGIN
 IF quelle + laenge < Rmem_Max THEN BEGIN
   Segm:=seg(ziel^);
   Offs:=ofs(ziel^);
   inc(Segm,Offs DIV 16);
   Offs:=Offs mod 16;
   inc(quelle,MByte1);
   mem_lesen(quelle,Offs,Segm,laenge);
  END ELSE BEGIN;
    ASM mov ax,0003; int 10h; END;
    writeln('Fehler beim Rcklesen des XMS.');
    writeln('System angehalten!');
    halt(0);
  END;
END;

PROCEDURE Rmem_write(quelle:pointer;ziel:longint;laenge:word);
{
 *************************************************************************
 ***                                                                   ***
 ***       Kopiert einen Block aus dem Hauptspeicher ins RMEM          ***
 ***                                                                   ***
 *************************************************************************
}
BEGIN
 IF ziel+laenge < Rmem_Max THEN BEGIN
   Segm := seg(quelle^);
   Offs := ofs(quelle^);
   inc(Segm,Offs DIV 16);
   Offs := Offs mod 16;
   inc(ziel,MByte1);
   mem_write(ziel, Offs,Segm,laenge);
  END ELSE BEGIN;
    ASM mov ax,0003; int 10h; END;
    writeln('Speicheranforderungsfehler. Nicht genug Speicher?');
    writeln('System angehalten!');
    halt(0);
  END;
END;

PROCEDURE memory_checks(minmain,minxms : word);
{
 *************************************************************************
 ***                                                                   ***
 ***         Prft, ob gengend Speicher zur Verfgung steht           ***
 ***                                                                   ***
 *************************************************************************
}
VAR xmsfree,mainfree : word;
BEGIN;
  { Freien XMS - Speicher ermitteln }
  xmsfree := xms_free DIV 1024;
  { Hauptspeicher ermitteln }
  mainfree := memavail DIV 1024;
  { Meldung, wenn nicht genug ferier Speicher }
  IF (xmsfree < minxms) OR (mainfree < minmain) THEN BEGIN;
    ASM mov ax,0003; int 10h; END;
    writeln('Nicht gengend Speicher vorhanden!');
    writeln('           Bentigt:      Verfgbar:');
    writeln('XMS :     ',minxms :6,' KB         ',xmsfree:4,' KB');
    writeln('Konv.:    ',minmain:6,' KB         ',mainfree:4,' KB');
    halt(0);
  END;
END;

FUNCTION Rgetmem(VAR rpos : longint;rsize : longint) : boolean;
{
 *************************************************************************
 ***                                                                   ***
 ***        Eine vereinfachte Getmem-PROCEDURE fr das RMEM            ***
 ***                                                                   ***
 *************************************************************************
}
BEGIN;
  IF Rmemposi+rsize > Rmem_max THEN BEGIN;
    Rgetmem := false;
  END ELSE BEGIN;
    rpos := Rmemposi;
    inc(Rmemposi,rsize);
    Rgetmem := true;
  END;
END;

FUNCTION RFreeMem(rsize: longint): boolean;
{
 *************************************************************************
 ***                                                                   ***
 ***        Eine vereinfachte Freemem-PROCEDURE fr das RMEM           ***
 ***                                                                   ***
 *************************************************************************
}
BEGIN
     IF (Rmemposi-rsize) < Mbyte1 THEN rfreemem := false
        ELSE BEGIN
             dec(rmemposi,rsize);
             rfreemem := true;
        END;
END;

PROCEDURE Exit_Rmem;
{
 *************************************************************************
 ***                                                                   ***
 ***        Exit-PROCEDURE des RMEM, MUSS aufgerufen werden !          ***
 ***                                                                   ***
 *************************************************************************
}
BEGIN;
  { Block entsperren }
  XMS_unlock(My_XmsHandle);
  { Speicher freigeben }
  Freemem_XMS(My_XmsHandle);
END;

PROCEDURE enable_Realmem(Min : word);
{
 *************************************************************************
 ***                                                                   ***
 ***               Schaltet in den RMEM - Modus                        ***
 ***    Es mu "MIN" KB freier XMS-Speicher  vorhanden sein !          ***
 ***                                                                   ***
 *************************************************************************
}
BEGIN
 { Auf Multitasker prfen ... }
 IF multitasker_aktiv THEN BEGIN;
   ASM mov ax,0003; int 10h; END;
   writeln('Der Prozessor befindet sich schon im V86-Modus!');
   writeln('Bitte starten sie das System OHNE Memory-Manager wie z.B. QEMM, EMM386 etc.');
   writeln('HIMEM.SYS wird allerdings bentigt! ');
   halt(0);
 END;
 { XMS Treiber installiert ?  }
 IF NOT XMS_Vorhanden THEN BEGIN;
   ASM mov ax,0003; int 10h; END;
   writeln('Entweder kein XMS oder HIMEM.SYS verfgbar.');
   writeln('Bitte starten Sie das System MIT HIMEM.SYS!');
   halt(0);
 END;
 { Bentigten Speicher belegen }
 error := Getmem_XMS(My_XmsHandle,min*1024);
 IF error <> 0 THEN BEGIN;
   ASM mov ax,0003; int 10h; END;
   writeln('Fehler bei der Speicherbelegung!');
   writeln('Es werden min. ',Min,' KB freier XMS-Speicher bentigt!!');
   writeln('Bitte starten Sie das System MIT HIMEM.SYS!');
   writeln;
   halt(0);
 END;
 { Physikalische Startposition ermitteln }
 Rmemposi := XMS_lock(My_XmsHandle);
 IF rmemposi < 1000000 THEN BEGIN;
   ASM mov ax,0003; int 10h; END;
   writeln('Fehler bei der Speicherfixierung!');
   writeln('Es werden min. ',Min,' KB freier XMS-Speicher bentigt!!');
   writeln('Bitte starten Sie das System MIT HIMEM.SYS!');
   writeln;
   halt(0);
 END;
 { Freischalten }
 Enable_4Giga;
END;

BEGIN;
  Check_for_XMS;
  Rmem_Max := XMS_Free;
END.
