UNIT VOCs; {$M 4000,12000,12000}

{

  * UNIT ***** VOCs ******************************************************
  * Von ****** Jan Strter ***********************************************
  Erstellt am:             09.11.1996
  Letzte nderung am:      14.04.1997

  Zweck:                   Tool zur Ausgabe von VOC-Dateien ber SB-Karte

  * Copyright-Hinweis ****************************************************
  Diese kleine Unit wurde vom Autor als FreeWare deklariert, darf also
  kostenlos kopiert und verwendet werden.

  * Kurze Beschreibung der Funktionsweise ********************************
  Die Unit mu ersteinmal ber die Funktion VOC_INIT initialisiert werden,
  nachdem das passiert ist, kann man die Voc-Dateien laden. Die Daten
  landen dann im XMS, eine Mglichkeit, die Daten in den konventionellen
  Speicher zu laden, ist nicht vorgesehen, da die meisten Voc-Dateien
  sowieso grer als 64KB sind und nicht mehr in ein Segment passen.
  Dazu mu die Prozedur VOC_LOAD benutzt werden, die einen Record fr
  die Sounddaten und den Namen der Datei als Parameter erwartet.
  Es gibt einige Einschrnkungen, die die Vocs erfllen mssen, soll
  alles korrekt ablaufen:
        1) alle Voc-Dateien mssen die GLEICHE SAMPLERATE haben
        2) alle Datein mssen MONO und 8-BITTIG sein.
  Das klingt erstmal nach ziemlichen Handicaps, aber mehr braucht man
  in den meisten Fllen sowieso nicht, darum habe ich aus
  Geschwindigkeitsgrnden darauf verzichtet, eine Untersttzung fr
  alle existierenden Voc-Formate zu programmieren. Bei der Entstehung
  der Unit standen nmlich nicht Kompatiblitt, sondern Geschwindigkeit
  und Krze des Source-Codes im Vordergrund. Die ganze Unit ist krzer
  als 1000 Zeilen, also nicht sehr lang, und darauf kam es mir an.
  Aber zurck zum Thema!
  Nachdem die Samples im Speicher sind, wird es Zeit, sie abzuspielen.
  Dazu gibt es die Prozedur VOC_PLAY, der man einen Record fr die
  Sampledaten, die Prioritt des Samples, die Lautsrke und schlielich
  eine Nummer zwischen 0 und 3 bergibt. Der letzte Parameter gibt den
  Kanal an, auf dem der Sample ausgegeben werden soll, ist der Wert
  grer als 3, dann sucht sich die Unit selbst einen freien Kanal
  raus und verwendet diesen. Wenn schon alle Kanle belegt sind,
  kommt es auf die Prioritt an, welcher Sample gespielt wird und welcher
  vernachlssigt wird. Durch dieses System kann verhindert werden, da
  wichtige Sounds von anderen unterbrochen werden.
  Das war dann eigentlich auch schon alles Wichtige. Es gibt noch einige
  weitere Prozeduren, die weiter unten erlutert werden.

}

{ ************************************************************************ }
INTERFACE
{ ************************************************************************ }

USES dos;

TYPE
     XMSHandle = word;

     vocheader = RECORD
       scanstr : ARRAY[0..19] OF char;
       Sampoff : word;
       Verslo  : Byte;
       Vershi  : Byte;
       scan    : word;
     END;

     Voiceblock = RECORD
       scan     : byte;
       Length_lo: word;
       Length_hi: byte;
       SR       : byte;
       Pack     : byte;
     END;

     voc_Type_internal = RECORD
       Handle : xmshandle;
       Size   : longint;
       SRate  : longint;
       act_pos: longint;
       stereo : boolean;
       Volume : byte;
       Prior  : byte;
       Stop   : boolean;
       Loop   : boolean;
     END;

     voc_Type = RECORD
       Handle : xmshandle;
       Size   : longint;
       SRate  : longint;
       stereo : boolean;
     END;

CONST
     voc_no_priority = 0;
     voc_high_priority = 255;
     voc_any_chn = 255;
     voc_chn_1 = 0;
     voc_chn_2 = 1;
     voc_chn_3 = 2;
     voc_chn_4 = 3;
     voc_noerror = 0;
     voc_OutOfMemError = 1;
     voc_IOError = 2;
     voc_AnyError = 3;
     voc_error: byte = voc_noerror;


FUNCTION voc_reset_blaster: boolean;
{
  Resettet die SoundBlaster-Karte und gibt zurck, ob der Reset
  erfolgreich war, oder nicht.
}
PROCEDURE voc_set_volume(vol: byte);
{
  Setzt die Master-Lautstrke, mit der die Samples ausgegeben werden.
  Aus ist 0, volle Pulle 64.
}
PROCEDURE voc_set_sample_rate(sr: byte);
{
  Setzte die Sample-Rate, mit der die Samples ausgegeben werden.
  Die Formel lautet: 256 - (1000000 div Samplerate)
  Beispiel: voc_set_volume( BYTE( 256 - LONGINT(1000000 DIV 11000)));
}
PROCEDURE voc_load(filename: STRING; VAR zvoc: voc_type);
{
  Ldt die Voc-Datei mit dem Namen FILENAME in den Sound-Record ZVOC.
  Der Fehler-Code wird in VOC_ERROR zurckgegeben.
}
PROCEDURE voc_free(VAR zvoc: voc_type);
{
  Gibt den durch ZVOC belegten XMS-Speicher wieder frei.
  ACHTUNG: DIESE PROZEDUR MUSS AM ENDE DES PROGRAMMS AUFGERUFEN WERDEN,
  DA DER BELEGTE XMS-SPEICHER NICHT AUTOMATISCH WIEDER FREIGEGEBEN WIRD!
}
FUNCTION voc_init(base: word; irq,dma: byte): boolean;
{
  Initialisiert die Unit und die SB-Karte mit den bergeben Werten.
  Bsp.: VOC_INIT($220,5,1);
  (Port 220h, IRQ 5, DMA 1)
}
PROCEDURE voc_stop(channel: byte);
{
  Stoppt die Ausgabe des Kanals CHANNEL.
}
PROCEDURE voc_stop_all;
{
  Stoppt alle Kanle.
}
PROCEDURE voc_pause;
{
  Hlt die Ausgabe an.
}
PROCEDURE voc_continue;
{
  Setzte die Ausgabe fort.
}
PROCEDURE voc_play(qvoc: voc_type; pr: byte; vol,chn: byte; looped: boolean);
{
  Spielt QVOC mit der Prioritt PR und der Lautsrke VOL auf
  dem Kanal CHN ab.
  Die maximale Lautsrke ist 64. Wenn CHN grer 3 ist, dann sucht
  sich die Unit einen freien Kanal oder einen Kanal mit niedrigerer
  Prioritt heraus.
  Falls LOOPED true ist, wird der Sound automatisch wiederholt (geloopt).
}
FUNCTION voc_is_playing(qvoc: voc_type): integer;
{
  Gibt zurck, ob der Sample QVOC gerade aktiv ist, oder nicht.
}

{ ************************************************************************ }
IMPLEMENTATION
{ ************************************************************************ }

TYPE
    pt = RECORD
         ofs,seg: word;
    END;

    XMS_Copyblock = RECORD
       Size     : longint;
       Q_Handle : Word;
       Q_Offset : pointer;
       Z_Handle : Word;
       Z_Offset : pointer;
    END;

CONST
     SBRegDetected: boolean = false;
     MixerDetected: boolean = false;
     Stereo: boolean = false;
     dsp_irq: byte = 5;
     dma_ch: byte = 1;
     dsp_adr: word = $220;
     bufsize = 500;
     act_buf: byte = 0;
     vocs_initialized: boolean = false;
     maxvoc = 3;
     interrupt_running: boolean = false;
     sample_rate: byte = 166;
     volume: byte = 64;
     voc_buf_size = 5000;
     buffers_size = bufsize;
     pause: boolean = false;
     dma_page: ARRAY[0..3] OF byte = ($87,$83,$81,$81);
     dma_adr: ARRAY[0..3] OF byte = (0,2,4,6);
     dma_wc: ARRAY[0..3] OF byte = (1,3,5,7);

VAR
     XMS_Vorhanden: boolean;
     XMST: pointer;
     XMS_Version: word;
     XC: XMS_Copyblock;
     a: byte;
     offsets: ARRAY[0..maxvoc] OF word;
     segments: ARRAY[0..maxvoc] OF word;
     buffers: ARRAY[0..maxvoc] OF pointer;
     sounds: ARRAY[0..maxvoc] OF voc_type_internal;
     oldexitproc: pointer;
     oldint: pointer;
     buf: ARRAY[0..1] OF pointer;
     irqmask: byte;
     voc_buf: pointer;


{ ************************************************************************
  Erstmal die XMS-Prozeduren und Funtionen
  ************************************************************************ }

PROCEDURE Check_for_XMS; ASSEMBLER;
ASM
    mov ax,4300h
    int 2Fh
    cmp al,80h
    jne @Kein_XMSTreiber
    mov ax,4310h
    int 2Fh
    mov word ptr XMST + 2,es
    mov word ptr XMST + 0,bx
    xor ax,ax
    call dword ptr [XMST]
    cmp  ax,0200h
    jb   @Kein_XMSTreiber
    mov  XMS_Version,ax
    mov  XMS_Vorhanden,0
@Kein_XMSTreiber:
    mov XMS_Vorhanden,1
@Ende_XMS_Check:
END;

FUNCTION XMS_2_Ram(d : pointer; h : XMSHandle; Size : Word; XMSOffset : longint) : byte;
VAR fresult : byte;
BEGIN
  XC.Size     := Size;
  XC.Q_Handle := h;
  XC.Q_Offset := pointer(xmsoffset);
  XC.Z_Handle := 0;
  XC.Z_Offset := d;
  ASM
    mov si,offset XC
    mov ax,0B00h
    call dword ptr [XMST]
    cmp ax,1
    jne @Fehler_XMS2RAM
    mov fresult,0
    jmp @Ende_XMS2Ram
@Fehler_XMS2Ram:
    mov fresult,bl
@Ende_XMS2Ram:
  END;
  xms_2_ram:=fresult;
END;

FUNCTION RAM_2_XMS(q : pointer; h : XMSHandle; Size : Word; XMSOffset : longint) : byte;
VAR fresult : byte;
BEGIN
  XC.Size     := Size;
  XC.Q_Handle := 0;
  XC.Q_Offset := q;
  XC.Z_Handle := h;
  XC.Z_Offset := pointer(xmsoffset);
  ASM
    mov si,offset XC
    mov ax,0B00h
    call dword ptr [XMST]
    cmp ax,1
    jne @Fehler_RAM2XMS
    mov fresult,0
    jmp @Ende_Ram2XMS
@Fehler_Ram2XMS:
    mov fresult,bl
@Ende_Ram2XMS:
  END;
  ram_2_xms:=fresult;
END;

FUNCTION XMS_free  : longint;
VAR xms_in_kb : word;
    xms_long: longint;
BEGIN
  ASM
    mov ax,0800h
    call dword ptr [XMST]
    mov xms_in_kb,dx
  END;
  xms_long := xms_in_kb;
  XMS_free := xms_long * 1024;
END;

FUNCTION Getmem_XMS(VAR H : XMSHandle; Size : longint) : byte;
VAR bsize : word;
    Fresult : byte;
    xmsh : word;
BEGIN
  bsize := (size DIV 1024) + 1;
  ASM
    mov ax,0900h
    mov dx,bsize
    call dword ptr [XMST]
    cmp ax,1
    jne @Fehler_GetmemXms
    mov xmsh,dx
    mov Fresult,0
    jmp @Ende_GetmemXms
@Fehler_GetmemXMS:
    mov Fresult,bl
@Ende_GetmemXms:
  END;
  h := xmsh;
  Getmem_Xms := Fresult;
END;

FUNCTION Freemem_XMS(H : XMSHandle) : byte;
VAR fresult : byte;
BEGIN
  ASM
    mov ax,0a00h
    mov dx,h
    call dword ptr [XMST]
    cmp ax,1
    jne @Fehler_FreememXms
    mov Fresult,0
    jmp @Ende_FreememXms
@Fehler_FreememXms:
    mov Fresult,bl
@Ende_FreememXms:
  END;
  freemem_xms:=fresult;
END;

PROCEDURE XMS_Enable_A20; assembler;
ASM
  mov ax,0500h
  call dword ptr [XMST]
END;

PROCEDURE XMS_Disable_A20; assembler;
ASM
  mov ax,0600h
  call dword ptr [XMST]
END;

{ ************************************************************************
  Jetzt die wirklichen Sound-Routinen
  ************************************************************************ }

PROCEDURE Write_Mixer(Reg, Val: BYTE);
BEGIN
 Port[dsp_adr+$4] := Reg;
 Port[dsp_adr+$5] := Val;
END;


FUNCTION Read_Mixer(Reg: BYTE) : BYTE;
BEGIN
  Port[dsp_adr+$4] := Reg;
  Read_Mixer := Port[dsp_adr+$5];
END;

PROCEDURE reset_Mixer; ASSEMBLER;
ASM
  mov dx,dsp_adr+$4
  mov al,0
  out dx,al
  mov cx,50
@loop:
  loop @loop
  inc dx
  out dx,al
END;

FUNCTION Detect_Mixer_sb16 : BOOLEAN;
VAR SaveReg : WORD;
    NewReg  : WORD;
BEGIN
  Detect_Mixer_sb16 := MixerDetected;
  IF NOT SBRegDetected OR
  MixerDetected THEN EXIT;              { blaster-Karte vorhanden    }
                                        { oder Mixer-Chip schon      }
                                        { initalisiert               }
  Reset_Mixer;
  SaveReg := Read_Mixer($22);           { Register sichern           }
  Write_Mixer($22, 243);                { Wenn der geschribene wert  }
  NewReg  := Read_Mixer($22);           { mit dem zurckgelesenen    }
                                        { bereinstimmt, so ist ein  }
                                        { Zugriff mglich und somit  }
                                        { ein Mixer vorhanden        }
  IF NewReg = 243 THEN BEGIN
    MixerDetected := TRUE;
    STEREO := True;
  END;
  Write_Mixer($22, SaveReg);            { Altes Register zurck      }
  Detect_Mixer_sb16 := MixerDetected;
END;

PROCEDURE voc_set_volume;
BEGIN
     IF vol > 64 THEN vol := 64;
     volume := vol;
END;

PROCEDURE voc_set_sample_rate;
BEGIN
     sample_rate := sr;
END;

PROCEDURE write_dsp(v : byte);
BEGIN
  WHILE port[dsp_adr+$c] >= 128 DO;
  port[dsp_adr+$c] := v;
END;

FUNCTION read_sb: byte;
BEGIN
     WHILE port[dsp_adr+$a] = $AA DO;
     read_sb := port[dsp_adr+$a];
END;

{$F+}
PROCEDURE VOCs_exit;
BEGIN
     ExitProc := oldexitproc;
     setintvec($8+dsp_irq,oldint);
     IF NOT vocs_initialized THEN exit;
     port[$21] := Port[$21] OR irqmask;
     port[dsp_adr+$c] := $d3;
     Port[$20] := $20;
     Write_dsp($D0);
END;
{$F-}

PROCEDURE set_sample_rate(rate: byte);
BEGIN
     write_dsp($40);
     write_dsp(rate);
END;

PROCEDURE clear_buf(b: pointer; value: byte); ASSEMBLER;
ASM
   push es
   les di,b
   mov cx,buffers_size/2
   mov al,value
   mov ah,al
   rep stosw
   pop es
END;

PROCEDURE mix_buffer(p: pointer);
VAR
   seg_buf,ofs_buf: word;
   a: word;
   copy_size: integer;
   ofs_0,ofs_1,ofs_2,ofs_3: word;
   seg_0,seg_1,seg_2,seg_3: word;
   vol_0,vol_1,vol_2,vol_3: word;
   interpol: integer;
   _ofs_buf: word;
{   stimmen: byte;
   _offsets: array[0..maxvoc] of word;
   _volumes: array[0..maxvoc] of word;
   einzelbyte: shortint;
   einzelword: integer;
   b: byte;
   buffer_ofs: word;}
BEGIN
     seg_buf := seg(p^);
     ofs_buf := ofs(p^);

     { Die Dinger hier werden fr den Assembler-Teil gebraucht, weil man }
     { man nicht so ohne weiteres auf die globalen Arrays zugreifen kann... }
     ofs_0:=offsets[0];
     ofs_1:=offsets[1];
     ofs_2:=offsets[2];
     ofs_3:=offsets[3];
     seg_0:=segments[0];
     seg_1:=segments[1];
     seg_2:=segments[2];
     seg_3:=segments[3];
     vol_0:=sounds[0].volume;
     vol_1:=sounds[1].volume;
     vol_2:=sounds[2].volume;
     vol_3:=sounds[3].volume;

     { Jetzt die Sample-Daten in die Buffer schreiben, wenn ntig COPY_SIZE anpassen }
{     stimmen := 0;}
     FOR a := 0 TO maxvoc DO BEGIN
{         _offsets[a] := offsets[a];}
         clear_buf(buffers[a],128);
         IF sounds[a].stop = false THEN BEGIN
{            inc(stimmen);}
            copy_size := buffers_size;
            IF sounds[a].size-sounds[a].act_pos <= buffers_size THEN BEGIN
               copy_size := sounds[a].size-sounds[a].act_pos;
               IF sounds[a].loop = false THEN
                  sounds[a].stop := true
               ELSE BEGIN
                  { Sound ist geloopt }
                  IF copy_size MOD 2 <> 0 THEN dec(copy_size);
                  IF pause = false THEN BEGIN
                     xms_2_ram(buffers[a],sounds[a].handle,copy_size,sounds[a].act_pos);
                     inc(sounds[a].act_pos,copy_size);
                  END;
                  copy_size := buffers_size-copy_size;
                  sounds[a].act_pos := 0;
                  IF pause = false THEN BEGIN
                     xms_2_ram(buffers[a],sounds[a].handle,copy_size,sounds[a].act_pos);
                     inc(sounds[a].act_pos,copy_size);
                  END;
               END;
            END;
            IF copy_size MOD 2 <> 0 THEN dec(copy_size);
            { Und Daten aus dem XMS kopieren... }
            IF pause = false THEN BEGIN
               xms_2_ram(buffers[a],sounds[a].handle,copy_size,sounds[a].act_pos);
               inc(sounds[a].act_pos,copy_size);
            END;
            IF sounds[a].act_pos >= sounds[a].size THEN sounds[a].stop := true;
         END;
     END;
     {
     IF stimmen < 1 THEN stimmen := 1;

     buffer_ofs := ofs_buf;
     FOR a := 1 TO buffers_size DO BEGIN
         interpol := 0;
         FOR b := 0 TO maxvoc DO BEGIN
             einzelbyte := mem[segments[b]: _offsets[b]]-128;
             einzelword := einzelbyte * sounds[b].volume;
             einzelbyte := einzelword DIV 64;
             inc(_offsets[b]);
             interpol := interpol + einzelbyte;
         END;
         interpol := interpol DIV stimmen;
         interpol := interpol * volume;
         interpol := interpol DIV 64;
         interpol := interpol + 128;
         mem[seg_buf: buffer_ofs] := byte(interpol);
         inc(buffer_ofs);
     END;
     }

     { Und jetzt - TATAAAA - die Daten interpolieren... }
     _ofs_buf:=ofs_buf;
     ASM
        push es                         { ES sichern }

        mov cx,buffers_size             { BUFFERS_SIZE mal durchlaufen }

     @anfang:

        mov interpol,0                  { Werte auf Null }
        mov al,0
        mov ah,al

        mov es,seg_0                    { Adresse des Buffers laden }
        mov di,ofs_0
        mov al,es:[di]                  { Sample-Byte laden }
        mul vol_0                       { An Lautstrke anpassen }
        shr ax,6
        add interpol,ax                 { Zum Mix-Buffer addieren }
        inc ofs_0                       { Und Zeiger weiter }

        mov es,seg_1                    { s.o. }
        mov di,ofs_1
        mov al,es:[di]
        mov ah,0
        mul vol_1
        shr ax,6
        add interpol,ax
        inc ofs_1

        mov es,seg_2                    { s.o. }
        mov di,ofs_2
        mov al,es:[di]
        mov ah,0
        mul vol_2
        shr ax,6
        add interpol,ax
        inc ofs_2

        mov es,seg_3                    { s.o. }
        mov di,ofs_3
        mov al,es:[di]
        mov ah,0
        mul vol_3
        shr ax,6
        add interpol,ax
        inc ofs_3

        mov ax,interpol                 { INTERPOL nach AX }
        shr ax,2                        { Durch Stimmen (4) teilen }
        mul volume                      { An Master-Volume anpassen }
        shr ax,6
        mov es,seg_buf                  { Adresse des Mix-Buffers laden }
        mov di,_ofs_buf
        mov es:[di],al                  { Byte in den Buffer schreiben }
        inc _ofs_buf                    { Zeiger weitersetzen }

        dec cx                          { Zhler (CX) runter }
        jne @anfang                     { Wenn CX nicht Null, dann nochmal }

        pop es                          { ES restaurieren }
     END;
END;                            { Tsch... }

PROCEDURE Spiele_Sb(Segm,Offs,dsize : word);
VAR li : word;
BEGIN
  port[$0A] := dma_ch+4;                { DMA-Kanal sperren          }
  Port[$0c] := 0;                       { Adresse des Puffers (blk)  }
  Port[$0B] := $49;                     { fr Soundausgabe           }
  Port[dma_adr[dma_ch]] := Lo(offs);    { an DMA-Controller          }
  Port[dma_adr[dma_ch]] := Hi(offs);
  Port[dma_wc[dma_ch]] := Lo(dsize-1);  { Gre des Blockes (block-  }
  Port[dma_wc[dma_ch]] := Hi(dsize-1);  { groesse) an DMA-Controller }
  Port[dma_page[dma_ch]] := Segm;
  Write_dsp($14);
  Write_dsp(Lo(dsize-1));                  { Gre des Blockes an       }
  Write_dsp(Hi(dsize-1));                  { den DSP                    }
  Port[$0A] := dma_ch;                  { DMA-Kanal freigeben        }
END;

PROCEDURE Spiele_SbPro(Segm,Offs,dgr,dsize : word);
{
 Diese Procedure spielt den ber Segm:Offs adressierten Block mit der
 Gre dsize ab. Es ist darauf zu achten, das der DMA-Controller NICHT
 Seitenbergreifend arbeiten kann ...
}
VAR li : word;
BEGIN
  port[$0A] := dma_ch+4;                { DMA-Kanal sperren          }
  Port[$0c] := 0;                       { Adresse des Puffers  }
  Port[$0B] := $48+dma_ch;                     { fr Soundausgabe           }
  Port[dma_adr[dma_ch]] := Lo(offs);    { an DMA-Controller          }
  Port[dma_adr[dma_ch]] := Hi(offs);
  Port[dma_wc[dma_ch]] := Lo(dgr-1);  { Gre des Blockes (block-  }
  Port[dma_wc[dma_ch]] := Hi(dgr-1);  { groesse) an DMA-Controller }
  Port[dma_page[dma_ch]] := Segm;

  write_dsp($48);
  write_dsp(Lo(dsize-1));             { Gre des Blockes an       }
  write_dsp(Hi(dsize-1));             { den DSP                    }
  write_dsp($91);
  Port[$0A] := dma_ch;                  { DMA-Kanal freigeben        }
END;

PROCEDURE play_block_blaster(p: pointer; size: word);
VAR l : longint;
    pn,offs : word;
BEGIN
  l := 16*longint(pt(p).seg)+pt(p).ofs;
  pn := pt(l).seg;
  offs := pt(l).ofs;
  IF stereo THEN BEGIN
     spiele_sbpro(pn,offs,size,size);
 END  ELSE
     spiele_sb(pn,offs,size);
END;

PROCEDURE blaster_ready_int; INTERRUPT;
VAR
   h: byte;
   Save_Test8086: byte;
 BEGIN
     Save_Test8086 := Test8086;
     Test8086 := 0;
     IF NOT Pause THEN BEGIN
        h := port[dsp_adr+$E];
        set_sample_rate(sample_rate);
        IF act_buf = 0 THEN BEGIN
           play_block_blaster(buf[1],buffers_size);
           mix_buffer(buf[0]);
           act_buf := 1;
        END ELSE BEGIN
           play_block_blaster(buf[0],buffers_size);
           mix_buffer(buf[1]);
           act_buf := 0;
        END;
     END;
{     h:=Port[dsp_adr+$E];}
     Test8086 := Save_Test8086;
     port[$A0] := $20;
     port[$20] := $20;
END;

FUNCTION voc_reset_blaster;
VAR
   count, stat: byte;
BEGIN
     port[dsp_adr+$6] := 1;
     FOR count := 1 TO 100 DO;
     port[dsp_adr+$6] := 0;
     stat := 0;
     count := 0;
     WHILE (stat <> $AA) AND (count < 100) DO BEGIN
           stat := port[dsp_adr+$E];
           stat := port[dsp_adr+$A];
           inc(count);
     END;
     voc_reset_blaster := (stat = $AA);
END;

FUNCTION voc_init;
VAR
   a: byte;
BEGIN
     IF NOT xms_vorhanden THEN BEGIN
        voc_error := voc_OutOfMemError;
        voc_init := false;
        vocs_initialized := false;
        exit;
     END;
     dsp_adr := base;
     dsp_irq := irq;
     dma_ch := dma;
     IF voc_reset_blaster THEN BEGIN
        SBRegDetected := true;
        stereo := detect_mixer_sb16;
        voc_init := true;
        FOR a := 0 TO maxvoc DO BEGIN
            WITH sounds[a] DO BEGIN
                 handle := 0;
                 size := 0;
                 srate := 0;
                 act_pos := 0;
                 stereo := false;
                 prior := 0;
                 stop := true;
            END;
        END;
        set_sample_rate(256 - longint(1000000 DIV 11025));
        write_dsp($D1);
        setintvec(dsp_irq+$8,@blaster_ready_int);
        irqmask := 1 SHL dsp_irq;
        port[$21] := port[$21] AND NOT irqmask;
        act_buf := 1;
        mix_buffer(buf[0]);
        play_block_blaster(buf[0],buffers_size);
        vocs_initialized := true;
     END ELSE BEGIN
        voc_init := false;
        vocs_initialized := false;
     END;
END;

PROCEDURE voc_load;
CONST
     voc_scan_str = 'Creative Voice File'+#$1A;
VAR
   f: FILE;
   header: vocheader;
   block: voiceblock;
   inread: ARRAY[1..25] OF byte;
   SR: word;
   samplingr: word;
   gelesen: word;
   xms_ofs: longint;
   result: byte;
BEGIN
     IF pos('.',filename) = 0 THEN filename := filename + '.VOC';
     IF NOT vocs_initialized THEN BEGIN
        voc_error := voc_anyerror;
        exit;
     END;

     { Alles auf Null ! }
     fillchar(zvoc,sizeof(zvoc),0);

     { Erstmal prfen, ob Datei vorhanden, dann ffnen }
     assign(f,filename);
     {$I-} reset(f,1); {$I+}
     IF IOresult <> 0 THEN BEGIN
        voc_error := voc_IOerror; { Falscher Dateiname? }
        exit;
     END;

     { Header einlesen und prfen, ob gltig }
     zvoc.size := filesize(f)-sizeof(header)-sizeof(inread)-50;
     IF filesize(f) < sizeof(header)+sizeof(inread) THEN BEGIN
        close(f);
        voc_error := voc_anyerror; { zu klein }
     END;
     blockread(f,header,$19);
     IF header.scanstr <> voc_scan_str THEN BEGIN
        close(f);
        voc_error := voc_anyerror; { Keine VOC-Datei }
     END;

     { Speicher im XMS allokieren }
     IF xms_free < zvoc.size THEN BEGIN
        close(f);
        voc_error := voc_OutOfMemError;
        exit;
     END;
     result := getmem_xms(zvoc.handle,zvoc.size);
     IF result <> 0 THEN BEGIN
        voc_error := voc_outofmemerror;
        close(f);
        exit;
     END;

     { Jetzt VoiceBlock laden und Sample-Rate feststellen }
     { (dank Creative Labs gibt's ja tausende von verschiedenen VOC-Versionen...) }
     blockread(f,inread,20);
     block.scan := inread[2];
     IF block.scan = 1 THEN BEGIN
        block.sr := inread[6];
     END;
     IF block.scan = 8 THEN BEGIN
        SR := inread[6]+(inread[7]*256);
        samplingr := 256000000 DIV (65536 - SR);
        IF inread[9] = 1 THEN zvoc.stereo := true;
        block.SR := 256 - longint(1000000 DIV samplingr);
     END;
     IF block.scan = 9 THEN BEGIN
        samplingr := inread[6]+(inread[7]*256);
        IF inread[11] = 2 THEN zvoc.stereo := true;
        block.SR := 256 - longint(1000000 DIV samplingr);
     END;
     zvoc.srate := block.SR;

     { Und jetzt: den ganzen Kram in den XMS laden... }
     xms_ofs:=0;
     REPEAT
           blockread(f,voc_buf^,voc_buf_size,gelesen);
           IF gelesen MOD 2 <> 0 THEN dec(gelesen);
           ram_2_xms(voc_buf,zvoc.handle,gelesen,xms_ofs);
           inc(xms_ofs,gelesen);
     UNTIL gelesen <> voc_buf_size;
     zvoc.size := xms_ofs-10;
     { Datei nicht mehr bentigt - schlieen ! }
     close(f);
END;

PROCEDURE voc_free;
BEGIN
     IF zvoc.handle <> 0 THEN BEGIN
        freemem_xms(zvoc.handle);
        fillchar(zvoc,sizeof(zvoc),0);
     END;
END;

PROCEDURE voc_stop;
BEGIN
     IF channel < 4 THEN sounds[channel].stop := true;
END;

PROCEDURE voc_stop_all;
VAR
   a: byte;
BEGIN
     FOR a := 0 TO maxvoc DO sounds[a].stop := true;
END;

PROCEDURE voc_pause;
BEGIN
     pause := true;
END;

PROCEDURE voc_continue;
VAR
   h: byte;
BEGIN
     pause := false;
     IF NOT vocs_initialized THEN exit;
     h := port[dsp_adr+$E];
     set_sample_rate(sample_rate);
     IF act_buf = 0 THEN BEGIN
        play_block_blaster(buf[1],buffers_size);
        mix_buffer(buf[0]);
        act_buf := 1;
     END ELSE BEGIN
        play_block_blaster(buf[0],buffers_size);
        mix_buffer(buf[1]);
        act_buf := 0;
     END;
END;

PROCEDURE voc_play;
VAR
   a: byte;
BEGIN
     { ggf. falsche Werte berichtigen }
     IF vol > 64 THEN vol := 64;
     { Wenn Kanal angegeben wurde, diesen prfen und ggf. berschreiben }
     IF chn < maxvoc+1 THEN BEGIN
        IF (sounds[chn].stop = true) OR (sounds[chn].prior <= pr) THEN BEGIN
           WITH sounds[chn] DO BEGIN
                prior := pr;
                volume := vol;
                handle := qvoc.handle;
                size := qvoc.size;
                act_pos := 0;
                stereo := qvoc.stereo;
                stop := false;
                srate := qvoc.srate;
                loop := looped;
                exit;
           END;
        END;
     END;
     { Erstmal gucken, ob ein freier Kanal }
     FOR a:=0 TO maxvoc DO BEGIN
         IF sounds[a].stop = true THEN BEGIN
            WITH sounds[a] DO BEGIN
                 prior := pr;
                 volume := vol;
                 handle := qvoc.handle;
                 size := qvoc.size;
                 act_pos := 0;
                 stereo := qvoc.stereo;
                 stop := false;
                 srate := qvoc.srate;
                 loop := looped;
                 exit;
            END;
         END;
     END;
     { Kein Kanal frei, also einen mit niedrigerer Prioritt suchen }
     FOR a:=0 TO maxvoc DO BEGIN
         IF sounds[a].prior < pr THEN BEGIN
            WITH sounds[a] DO BEGIN
                 prior := pr;
                 volume := vol;
                 handle := qvoc.handle;
                 size := qvoc.size;
                 act_pos := 0;
                 stereo := qvoc.stereo;
                 stop := false;
                 srate := qvoc.srate;
                 loop := looped;
                 exit;
            END;
         END;
     END;
     { Keiner mit niedrigerer Prioritt, also einen mit gleicher suchen }
     FOR a:=0 TO maxvoc DO BEGIN
         IF sounds[a].prior = pr THEN BEGIN
            WITH sounds[a] DO BEGIN
                 prior := pr;
                 volume := vol;
                 handle := qvoc.handle;
                 size := qvoc.size;
                 act_pos := 0;
                 stereo := qvoc.stereo;
                 stop := false;
                 srate := qvoc.srate;
                 loop := looped;
                 exit;
            END;
         END;
     END;
     { Wenn das Programm hier landet, konnte der Sound nicht gespielt werden. }
END;

FUNCTION voc_is_playing;
VAR
   a: byte;
BEGIN
     voc_is_playing := -1;
     FOR a := 0 TO maxvoc DO BEGIN
         IF (sounds[a].handle = qvoc.handle) AND (sounds[a].stop = false) THEN
            voc_is_playing := a;
     END;
END;

BEGIN
     check_for_xms;
     getintvec($8+dsp_irq,oldint);
     oldexitproc := exitproc;
     exitproc := @vocs_exit;
     getmem(voc_buf,voc_buf_size);
     getmem(buf[0],bufsize);
     getmem(buf[1],bufsize);
     FOR a:=0 TO maxvoc DO BEGIN
         getmem(buffers[a],buffers_size);
         offsets[a]:=ofs(buffers[a]^);
         segments[a]:=seg(buffers[a]^);
     END;
END.